from statistics import mean
from bisect import bisect_left


def getVowelKey(s, vList):
    uniq = set(s)
    vow = [v for v in uniq if v in vList]
    cNum = len(uniq) - len(vow)
    vowX = vow if len(vow) else "x"
    return "".join([str(cNum)] + sorted(vowX))


def toLetterMap(scoreMap, vList):
    letterMap = {}
    kScoreMap = {}
    # Insure highest score combos first
    for combo, score in scoreMap.items():
        key = getVowelKey(combo, vList)
        combos = letterMap.get(key, [])
        scores = kScoreMap.get(key, [])
        newIndex = bisect_left(scores, -score)
        scores.insert(newIndex, -score)
        combos.insert(newIndex, combo)
        letterMap[key] = combos
        kScoreMap[key] = scores
    letterMap[""] = []
    return letterMap


class Maps():

    def __init__(self, wsMap, vList):
        self.wsMap = wsMap
        self.wordMap = {}
        self.scoreMap = {}
        # Insure highest score words first
        for k, ws in wsMap.items():
            words = [w for (w, s) in ws]
            wsk = ({w: -s for (w, s) in ws}).get
            self.wordMap[k] = sorted(words, key=wsk)
            scores = [s for (w, s) in ws] or [0]
            self.scoreMap[k] = mean(scores)

        self.letterMap = toLetterMap(self.scoreMap, vList)


def mapFunction(config):
    def toMaps(wsMap):
        vList = config.vList
        return Maps(wsMap, vList)
    return toMaps



import logging

# https://link.springer.com/content/pdf/10.3758/BF03195586.pdf
PRIORS = {
    "e": 0.1255,
    "t": 0.0893,
    "a": 0.0853,
    "o": 0.0767,
    "n": 0.0735,
    "i": 0.0734,
    "s": 0.0679,
    "r": 0.0671,
    "h": 0.0479,
    "l": 0.0414,
    "d": 0.0384,
    "c": 0.0318,
    "u": 0.0262,
    "m": 0.0238,
    "f": 0.021,
    "p": 0.0204,
    "g": 0.0196,
    "y": 0.0172,
    "w": 0.0165,
    "b": 0.014,
    "v": 0.0106,
    "k": 0.0075,
    "x": 0.002,
    "z": 0.0011,
    "j": 0.0011,
    "q": 0.0009
}
MAXIMA = [0, 128, 512, 1024]


class Defaults():

    priors = PRIORS
    oDir = "results"
    eDir = "effects"
    maxRatio = 0.01
    maxCounts = {
        1: MAXIMA[1-1],
        2: MAXIMA[2-1],
        3: MAXIMA[3-1],
        "*": MAXIMA[-1]
    }
    maxReps = {
        "1x": 0,
        "2x": 1,
        "3x": 1,
        "4x": 0,
        "5x": 0,
        "6x": 0,
    }
    minWords = 4
    maxWords = 7
    maxConsonants = 6
    maxVowels = 2


def normalizePriors(priors):
    n = len(priors)
    precision = 100
    priorSum = sum(priors.values()) or 1
    if n > 0 and int(priorSum * precision) != precision:
        s = f"as sum of {priorSum} != 1.0"
        logging.warning(f"Normalizing {n} prior items {s}")
    return {k: v/priorSum for k, v in priors.items()}


def normalizeArguments(*args):
    (maxRatio, maxCounts, maxReps, priors) = args[:4]
    (minWords, maxWords, maxC, maxV) = args[4:]
    # Maximum ratio of n-letter words found
    maxRatio = maxRatio if maxRatio else Defaults.maxRatio
    # Max number of n-letter words
    maxCounts = {**Defaults.maxCounts, **maxCounts}
    # Limit words without vowels
    maxReps = {**Defaults.maxReps, **maxReps}
    # Normalized letter frequency probability distributions
    priors = normalizePriors({**Defaults.priors, **priors})

    return [
        maxRatio,
        maxCounts,
        maxReps,
        priors,
        minWords if minWords else Defaults.minWords,
        maxWords if maxWords else Defaults.maxWords,
        maxC if maxC else Defaults.maxConsonants,
        maxV if maxV else Defaults.maxVowels
    ]



from .defaults import Defaults, MAXIMA
from .logHandler import makeHandler

import logging
import argparse
from pathlib import Path
from datetime import datetime

FMT = argparse.ArgumentDefaultsHelpFormatter


def parseMaxima(arguments):
    maxima = enumerate(arguments.maxima)
    maxCounts = {i+1: c for i, c in maxima}
    if len(maxCounts):
        maxCounts["*"] = max(maxCounts.values())
    return maxCounts


def toLogFile(ext):
    fmt = f'%Y-%m-%dT%H%M.{ext}'
    return datetime.today().strftime(fmt)


class Text():

    tagline = "Find perfect pangrams."
    shh = "Write less, not to log file"
    debug = "Increase verbosity of logs"
    showAll = "Show all redundant pangrams"
    results = "Folder for output pangrams"
    effects = "Folder for intermediate steps"
    maxima = "Set max number of words for each length"
    maximumRatio = "Use at most this ratio of found words"
    maxConsonants = "Maximum consonants per word"
    maxVowels = "Maximum vowels per word"
    minWords = "Minimum words per sentence"
    maxWords = "Maximum words per sentence"


class Arguments():

    argumentKeys = {
        "formatter_class": FMT,
        "description": Text.tagline
    }
    constants = [
        Defaults.maxReps,
        Defaults.priors
    ]
    max_ = {
        "default": MAXIMA,
        "help": Text.maxima,
        "metavar": "N",
        "nargs": "*",
        "type": int
    }
    a_ = {
        "default": False,
        "help": Text.showAll,
        "action": "store_true"
    }
    s_ = {
        "default": False,
        "help": Text.shh,
        "action": "store_true"
    }
    d_ = {
        "default": False,
        "help": Text.debug,
        "action": "store_true"
    }
    o_ = {
        "default": Defaults.oDir,
        "metavar": Defaults.oDir,
        "help": Text.results,
        "type": Path
    }
    e_ = {
        "default": Defaults.eDir,
        "metavar": Defaults.eDir,
        "help": Text.effects,
        "type": Path
    }
    r_ = {
        "default": Defaults.maxRatio,
        "help": Text.maximumRatio,
        "metavar": "R+",
        "type": float
    }
    w_ = {
        "default": Defaults.minWords,
        "help": Text.minWords,
        "metavar": "W-",
        "type": int
    }
    ww_ = {
        "default": Defaults.maxWords,
        "help": Text.maxWords,
        "metavar": "W+",
        "type": int
    }
    c_ = {
        "default": Defaults.maxConsonants,
        "help": Text.maxConsonants,
        "metavar": "C+",
        "type": int
    }
    v_ = {
        "default": Defaults.maxVowels,
        "help": Text.maxVowels,
        "metavar": "V+",
        "type": int
    }

    def __init__(self):

        dFlags = ('--debug', '-d')
        sFlags = ('--shh', '-s')
        aFlags = ('--all', '-a')
        arg_ = self.argumentKeys
        parser = argparse.ArgumentParser(**arg_)
        parser.add_argument('maxima', **self.max_)
        parser.add_argument(*aFlags, **self.a_)
        parser.add_argument(*dFlags, **self.d_)
        parser.add_argument(*sFlags, **self.s_)
        parser.add_argument('-o', **self.o_)
        parser.add_argument('-e', **self.e_)
        parser.add_argument('-r', **self.r_)
        parser.add_argument('-m', **self.w_)
        parser.add_argument('-w', **self.ww_)
        parser.add_argument('-c', **self.c_)
        parser.add_argument('-v', **self.v_)
        self.parser = parser

    @staticmethod
    def needed():
        cls = Arguments
        n = len(cls.constants)
        for attribute in cls.__dict__.keys():
            isOwn = attribute[:2] != "__"
            isK = attribute[-1] == "_"
            if isOwn and isK:
                n += 1
        return n

    @property
    def list(self):
        parser = self.parser
        arguments = parser.parse_args()
        maxCounts = parseMaxima(arguments)
        argumentList = [
            arguments.debug,
            arguments.o,
            arguments.e,
            arguments.shh,
            arguments.all,
            arguments.r,
            maxCounts
        ]
        for value in self.constants:
            argumentList.append(value)
        argumentList.append(arguments.m)
        argumentList.append(arguments.w)
        argumentList.append(arguments.c)
        argumentList.append(arguments.v)
        return argumentList

    def setLogging(self, debug):
        filename = Path(toLogFile("log"))
        if filename.is_file():
            already = "Refusing to overwrite"
            raise OSError(f"{already} {filename}")
        filename.touch()
        handlers = [
            makeHandler(filename),
            makeHandler()
        ]
        level = "DEBUG" if debug else "INFO"
        log_ = {
            "encoding": "utf-8",
            "handlers": handlers,
            "level": getattr(logging, level)
        }
        logging.basicConfig(**log_)
        logging.info(f"Logging to {filename}")



from .readFromCache import readFromCache
from .readFromCache import writeLines
from .readFromCache import readLines
from .flushHandler import FlushHandler
from .patterns import patternFunction
from .maps import mapFunction

import logging
import pickle
import errno
import yaml
import sys
import os

noEnt = errno.ENOENT, os.strerror(errno.ENOENT)


def fileNotFound(filename):
    args = noEnt + (str(filename),)
    return FileNotFoundError(*args)


class FileHandler:

    def __init__(self, eDir, oDir, config):

        self.config = config
        toPath = self.config.toPath
        try:
            self.config.makeFolders([eDir, oDir])
        except ValueError as e:
            logging.error(e)
        self.patternFile = eDir / toPath("pattern", "yaml")
        self.wordFile = eDir / toPath("word", "yaml")
        self.emptyFile = eDir / toPath("empty", "txt")
        self.outFile = oDir / toPath("pangrams", "txt")
        self.sourceFile = eDir / "source.p"
        # Modify the config state
        flushEmpty = self.flushEmpty
        flushOutput = self.flushOutput
        self.config.flushHandler = FlushHandler(flushEmpty, flushOutput)

    def readEmptyFile(self):
        emptyFile = self.emptyFile
        if emptyFile.is_file():
            logging.info(f'Existing {emptyFile}')
            self.config.setEmpty(readLines(emptyFile))
            return self.config
        return self.config

    def writeLoop(self, source):
        sourceFile = self.sourceFile
        if sourceFile.is_file():
            logging.info(f'Existing {sourceFile}')
            return
        logging.info(f'Writing {sourceFile}')
        with open(sourceFile, "wb") as wf:
            pickle.dump(source, wf)

    def readLoop(self):
        sourceFile = self.sourceFile
        if not sourceFile.is_file():
            raise fileNotFound(sourceFile)
        with open(sourceFile, "rb") as rf:
            return pickle.load(rf)

    def writeMaps(self, maps):
        wordFile = self.wordFile
        wsMap = maps.wsMap
        if wordFile.is_file():
            logging.info(f'Existing {wordFile}')
            return
        logging.info(f'Writing {wordFile}')
        with open(wordFile, "w") as wf:
            yaml.dump(wsMap, wf)

    def readMaps(self):
        wordFile = self.wordFile
        if not wordFile.is_file():
            raise fileNotFound(wordFile)
        toMaps = mapFunction(self.config)
        return readFromCache(wordFile, toMaps)

    def writePatterns(self, patterns):
        tree = patterns.tree
        patternFile = self.patternFile
        if patternFile.is_file():
            logging.info(f'Existing {patternFile}')
            return
        logging.info(f'Writing {patternFile}')
        with open(patternFile, "w") as wf:
            yaml.dump(tree, wf)

    def readPatterns(self):
        patternFile = self.patternFile
        if not patternFile.is_file():
            raise fileNotFound(patternFile)
        empty = self.config.empty
        toPatterns = patternFunction(empty)
        return readFromCache(patternFile, toPatterns)

    def exit(self):
        try:
            self.finish(self.config, done=False)
        except (BrokenPipeError, IOError):
            pass
        sys.exit(0)

    def finishLoop(self, config):
        source = config.source
        self.writeLoop(source)
        sText = f"{len(source)} source words"
        logging.info(f'Saved {sText} to {self.sourceFile}')

    def flushOutput(self, config):
        self.config = config
        output = config.output
        writeLines(self.outFile, output)
        oText = f"{len(output)} perfect pangram sets"
        logging.info(f'Saved {oText} to {self.outFile}')

    def flushEmpty(self, config):
        self.config = config
        empty = config.empty
        writeLines(self.emptyFile, empty)
        eText = f"{len(empty)} unused patterns"
        logging.info(f'Saved {eText} to {self.emptyFile}')

    def finish(self, config, done=True):
        self.flushEmpty(config)
        self.flushOutput(config)
        if not done:
            logging.warning("Incomplete!")
        else:
            logging.info("Complete!")



from .patterns import patternFunction

from sympy.solvers.diophantine.diophantine import partition
from itertools import permutations, combinations
import logging


def getDict(d, keys):
    outRoot = d
    out = outRoot
    for k in keys:
        out[k] = out.get(k, {})
        out = out[k]
    return outRoot


def setDict(d, keys):
    if not len(keys):
        return d
    return getDict(d, keys)


def sortDict(d, keys, leaf):
    if not len(d):
        return ' '.join(leaf)
    return [sortDict(d[k], keys, [*leaf, k]) for k in keys if k in d]


def toKeyTuple(line):
    vcList = zip(line["v"], line["c"])
    joined = [''.join(map(str, vc[::-1])) for vc in vcList]
    return tuple(sorted(joined, reverse=True))


def noDecrease(pairList, keys):
    for key in keys:
        matchList = [b for a, b in pairList if a == key]
        for a, b in zip(matchList, matchList[1:]):
            if a > b:
                return False
    return True


def countRepeats(tup):
    return [(t, tup[:i].count(t)) for i, t in enumerate(tup)]


def findVowelLists(vow, cons):
    vowKeys = sorted(set(vow))
    vowOrder = countRepeats(vow)
    vowCopies = [v for v in set(vow) if vow.count(v) > 1]
    for perm in permutations(vowOrder):
        # If no-vowel word occurs twice, use only one permutation
        if not noDecrease(perm, vowCopies):
            continue
        vowOut = tuple([v for v, r in perm])
        vowIndices = [vowKeys.index(v) for v in vowOut]
        # For consonant counts repeated, use only one permuation
        if not noDecrease(list(zip(cons, vowIndices)), cons):
            continue
        yield vowOut


def splitPartsWithLimit(maxPart, numTotal, num):
    for part in partition(numTotal, num):
        if part[-1] > maxPart:
            continue
        yield part


def toConsonantCounts(config):
    maxConsonants = config.maxConsonants
    cCount = len(config.cList)
    for nWords in config.wordRange():
        splitArgs = (maxConsonants, cCount, nWords)
        for part in splitPartsWithLimit(*splitArgs):
            yield part


def joinSorted(x):
    return (''.join(sorted(x)),)


def groupByPartition(pool, part):
    if not len(part):
        yield ()
        return
    for chosen in combinations(pool, part[0]):
        nextPart = part[1:]
        nextPool = pool - set(chosen)
        for rest in groupByPartition(nextPool, nextPart):
            yield joinSorted(chosen) + rest


def toVowelLists(config):
    maxVowels = config.maxVowels
    vList = config.vList
    vCount = len(vList)
    uniq = set()
    for nWords in config.wordRange():
        maxNoVowel = int(nWords - vCount / maxVowels)
        for noVowelCount in range(maxNoVowel + 1):
            nVow = nWords - noVowelCount
            splitArgs = (maxVowels, vCount, nVow)
            for part in splitPartsWithLimit(*splitArgs):
                for vCombo in groupByPartition(set(vList), part):
                    noVowelTuple = ('x',) * noVowelCount
                    out = tuple(sorted(noVowelTuple + vCombo))
                    if out not in uniq:
                        uniq.add(out)
                        yield out


def toPatterns(config):
    vcMap = {}
    for nWords in config.wordRange():
        vcMap[nWords] = {'vow': [], 'cons': []}
    for v in toVowelLists(config):
        vcMap[len(v)]["vow"].append(v)
    for c in toConsonantCounts(config):
        vcMap[len(c)]["cons"].append(c)

    total = len(vcMap.items())
    for i, kd in enumerate(vcMap.items()):
        k, d = kd
        for vow in d["vow"]:
            for cons in d["cons"]:
                for v in findVowelLists(vow, cons):
                    yield {"len": k, "v": v, "c": cons}
        logging.info(f"{k}-word patterns ({i + 1}/{total})")


def invalidate(maps, config, vcl):
    letterMap = maps.letterMap
    maxReps = config.maxReps
    repMap = {}
    for k in vcl:
        if k not in letterMap:
            return True
        if k in maxReps:
            repMap[k] = repMap.get(k, 0) + 1
            if repMap[k] > maxReps[k]:
                return True
    return False


def listPatterns(maps, config):
    patterns = list(toPatterns(config))
    vcLists = [toKeyTuple(line) for line in patterns]
    vcKeys = set([vc for vcs in vcLists for vc in vcs])
    vcKeysDescending = sorted(vcKeys, reverse=True)
    wordTree = {}
    already = set()
    for vcl in vcLists:
        invalid = invalidate(maps, config, vcl)
        if vcl in already or invalid:
            continue
        wordTree = setDict(wordTree, vcl)
        already.add(vcl)
    tree = sortDict(wordTree, vcKeysDescending, [])
    toPattern = patternFunction(config.empty)
    return toPattern(tree)



from .maps import mapFunction
from .maps import getVowelKey
from .logging import CutLogger
from .logging import DeltaLogger
from .logging import logMaxCounts

from bisect import insort
import math


def getPriors(priors):
    def getPrior(char):
        default = 1 / len(priors)
        return priors.get(char, default)
    return getPrior


class Word():

    def __init__(self, word, frequency, letters, priors):
        self.word = word
        self.letters = letters
        pList = list(map(getPriors(priors), letters))
        self.score = frequency / math.prod(pList)


def isGoodWord(letters, config):
    maxConsonants = config.maxConsonants
    maxVowels = config.maxVowels
    vList = config.vList
    vCount = sum(v in letters for v in vList)
    cCount = len(letters) - vCount
    if cCount > maxConsonants or vCount > maxVowels:
        return False
    return True


def updateWords(keyMap, key, w):
    noted = keyMap.get(key, [])
    score = w.score
    word = w.word
    try:
        copyIndex = next(i for i, v in enumerate(noted) if v.word == word)
        noted[copyIndex].score = max(noted[copyIndex].score, score)
    except StopIteration:
        insort(noted, w, key=lambda v: v.score)
    return noted


def isOver(keyCount, letters, config):
    nLetters = len(letters)
    maxCounts = config.maxCounts
    totalWords = maxCounts.get(nLetters, maxCounts['*'])
    keyMax = totalWords / config.vowelCombos()
    return keyCount > keyMax


def hashWord(word):
    letterList = [c for c in set(word) if c.isalpha()]
    return ''.join(sorted(letterList))


def listMaps(loop, typos, config):
    kWords = {}
    kTotals = {}
    shh = config.shh
    dLogger = DeltaLogger(shh)
    priors = config.priors
    maxRatio = config.maxRatio
    maxCounts = config.maxCounts
    if "*" not in maxCounts:
        raise ValueError(logMaxCounts(maxCounts))

    for (word, frequency) in loop:
        letters = hashWord(word)
        # Repeat letters are impossible
        if len(word) > len(letters):
            continue
        config.addSource([word, frequency])
        # Ensure word meets configured requirements
        if not isGoodWord(letters, config):
            continue
        if word in typos:
            continue
        # Insert word sorted by score
        key = getVowelKey(letters, config.vList)
        w = Word(word, frequency, letters, priors)
        kWords[key] = updateWords(kWords, key, w)
        kTotals[key] = kTotals.get(key, 0) + 1
        # Cut lowest-scoring words
        oArgs = (len(kWords[key]), letters, config)
        over = isOver(*oArgs)
        logWord = word
        if over:
            logWord = kWords[key].pop(0).word
        # Logging
        dLogger.log(over, logWord, word)

    wsMap = {}
    cutLogger = CutLogger(shh)

    # Higher frequencies first
    for key, wList in kWords.items():
        total = kTotals.get(key, 0)
        maxAllowed = max(int(total*maxRatio), 1)
        cutLogger.add(wList, maxAllowed)
        # Allow at most fixed ratio of words
        for w in wList[::-1][:maxAllowed]:
            wValue = [w.word, w.score]
            letterList = wsMap.get(w.letters, [])
            wsMap[w.letters] = letterList + [wValue]

    cutLogger.log(maxRatio)

    toMaps = mapFunction(config)
    return toMaps(wsMap)



import logging

FMT0 = "{message}"
FMT1 = "{levelname} - {message}"
HTTP = "Starting new HTTP connection (%d): %s:%s"
FMTK = {
    "style": "{"
}


def isNoEnd(record):
    endCode = ', '
    msg = str(record.msg)
    return any([msg[-2:] == endCode, msg == "."])


def isGoogleDebug(record):
    msg = str(record.msg)
    isDebug = record.levelno == 10
    return isDebug and any([
        msg == HTTP, msg[:7] == "%s://%s"
    ])


def updater(self, fmt0, fmt1, record):
    noEnd = isNoEnd(record)
    if isGoogleDebug(record):
        return False
    isHigh = record.levelno > 20
    self.setFormatter(fmt1 if isHigh else fmt0)
    if noEnd:
        self.terminator = ''
    else:
        self.terminator = '\n'
    return True


class StreamHandler(logging.StreamHandler):

    fmt0 = logging.Formatter(FMT0, **FMTK)
    fmt1 = logging.Formatter(FMT1, **FMTK)

    def emit(self, record):
        if updater(self, self.fmt0, self.fmt1, record):
            return super().emit(record)


class FileHandler(logging.FileHandler):

    fmt0 = logging.Formatter(FMT0, **FMTK)
    fmt1 = logging.Formatter(FMT1, **FMTK)

    def emit(self, record):
        if updater(self, self.fmt0, self.fmt1, record):
            return super().emit(record)


def makeHandler(filename=None):

    return FileHandler(filename) if filename else StreamHandler()



from .pangrams import pangramsCli
from .readWordRecords import readWordRecords
from .defaults import normalizeArguments
from .listPatterns import listPatterns
from .logHandler import StreamHandler
from .fileHandler import FileHandler
from .typo.loadTypos import loadTypos
from .toPangrams import toPangrams
from .listMaps import listMaps
from .defaults import Defaults
from .config import Config

__all__ = [
    'pangramsCli',
    'readWordRecords',
    'normalizeArguments',
    'listPatterns',
    'StreamHandler',
    'FileHandler',
    'loadTypos',
    'toPangrams',
    'listMaps',
    'Defaults',
    'Config'
]



from .readWordRecords import readWordRecords
from .defaults import normalizeArguments
from .listPatterns import listPatterns
from .logHandler import StreamHandler
from .fileHandler import FileHandler
from .typo.loadTypos import loadTypos
from .toPangrams import toPangrams
from .listMaps import listMaps
from .defaults import Defaults
from .config import Config

from pathlib import Path
import logging


def makeConfig(shh, showAll, *args):
    return Config(shh, showAll, *normalizeArguments(*args))


def loadLoop(files):
    try:
        loop = files.readLoop()
    except FileNotFoundError as e:
        logging.info("Loading word loop, ")
        logging.warning(e)
        loop = readWordRecords()
    return loop


def savePangrams(oDir, eDir, *configArguments):
    log_ = {
        "handlers": [StreamHandler()],
        "encoding": "utf-8",
        "level": "INFO",
    }
    if not logging.getLogger().hasHandlers():
        logging.basicConfig(**log_)

    oDir = oDir if oDir else Defaults.oDir
    eDir = eDir if eDir else Defaults.eDir

    oDir.mkdir(parents=True, exist_ok=True)
    eDir.mkdir(parents=True, exist_ok=True)

    config = makeConfig(*configArguments)
    thisFolder = Path(__file__).parent.absolute()
    typoDataFolder = thisFolder / Path("typo", "data")
    typos = loadTypos(typoDataFolder)
    files = FileHandler(eDir, oDir, config)
    config = files.readEmptyFile()
    # Try to read existing files
    try:
        maps = files.readMaps()
    except FileNotFoundError as e:
        logging.info("Loading word maps, ")
        logging.warning(e)
        loop = loadLoop(files)
        maps = listMaps(loop, typos, config)
        files.writeMaps(maps)
        files.finishLoop(config)
    try:
        patterns = files.readPatterns()
    except FileNotFoundError as e:
        logging.info("Loading word patterns, ")
        logging.warning(e)
        patterns = listPatterns(maps, config)
        files.writePatterns(patterns)

    # Generate the actual pangrams
    try:
        config = toPangrams(maps, patterns, config)
    except KeyboardInterrupt:
        files.exit()
    files.finish(config, done=True)



from abeci.savePangrams import savePangrams
from abeci.arguments import Arguments

import logging


def pangramsCli():
    arguments = Arguments()
    args = arguments.list
    nArgs, nNeeded = len(args), Arguments.needed()
    if nArgs == nNeeded:
        pangramArgs = args[1:]
        shh = pangramArgs[2]
        if not shh:
            try:
                arguments.setLogging(args[0])
            except OSError as e:
                print(str(e))
        savePangrams(*pangramArgs)
    else:
        logging.error(f"Only {nArgs} of {nNeeded} args")



import logging
import pickle
import yaml


def readLines(readableFile):
    logging.info(f"Reading {readableFile}")
    if not readableFile.is_file():
        return []
    with open(readableFile, 'r') as rf:
        return rf.read().splitlines()


def writeLines(readableFile, lines):
    mode = 'a' if readableFile.is_file() else 'w'
    with open(readableFile, mode) as wf:
        for line in lines:
            wf.write(f"{line}\n")


def noChange(value):
    return value


def readFromCache(yamlFile, transform=None):
    transform = transform if transform else noChange
    pickleFile = yamlFile.with_suffix('.p')
    pickled = pickleFile.is_file()
    if pickled:
        with open(pickleFile, 'rb') as rf:
            logging.info(f'Reading {pickleFile}...')
            return pickle.load(rf)

    with open(yamlFile, 'r') as rf:
        logging.info(f'Reading {yamlFile}...')
        transformed = transform(yaml.safe_load(rf))
        if not pickled:
            logging.info(f'Writing {pickleFile}')
            with open(pickleFile, "wb") as wf:
                pickle.dump(transformed, wf)
        return transformed



import logging
import math


def logMaxCounts(maxCounts):
    mk = list(maxCounts.keys())
    text = 'must include "*" key'
    keys = ','.join(map(str, mk))
    return f"Keys {keys} of maxCounts {text}"


def roughLog(v):
    return 2 ** int(math.log2(max(1, math.log(v))))


class DeltaLogger():

    def __init__(self, shh):
        self.lastFirstLetter = " "
        self.logFirsts = (" ", " ")
        self.deltas = {"+": 0, "-": 0}
        self.logLen = None if shh else 500
        self.colon = ": ..." if shh else ":"

    def log(self, over, logWord, word):
        sym = "-" if over else "+"
        self.deltas[sym] += 1

        if self.logLen is not None:
            if self.deltas[sym] >= self.logLen:
                self.deltas[sym] = 0
                logText = f"{sym}{logWord}, "
                logging.info(logText)

        fw = word[0]
        f0, f1 = self.logFirsts
        ff = self.lastFirstLetter

        if fw != ff and f0 == ff and fw == f1:
            self.lastFirstLetter = fw
            logging.info(f"\n{fw}{self.colon}")

        self.logFirsts = (f1, fw)


class CutLogger():

    def __init__(self, shh):
        self.wordCount = 0
        self.cutWordCount = 0
        self.cutWordLogs = []
        self.shh = shh

    def add(self, wList, maxAllowed):
        cutKeyCount = max(0, len(wList) - maxAllowed)
        self.cutWordCount += cutKeyCount
        self.wordCount += len(wList)
        if self.shh:
            return
        if cutKeyCount > 0:
            bestCut = wList[cutKeyCount - 1].word
            phrase = [str(cutKeyCount), "like", bestCut]
            self.cutWordLogs.append(" ".join(phrase))

    def log(self, maxRatio):
        wc = self.wordCount
        cwc = self.cutWordCount
        ratio = 100 * maxRatio
        logging.info("")
        logging.debug(wc, cwc)
        if cwc > 0:
            cut = int(1000 * cwc / wc) / 10
            cutLog = f"cut {cut}% of {wc} words"
            logging.info(f"The {ratio}% has {cutLog}:")
            if len(self.cutWordLogs):
                logging.info(", ".join(self.cutWordLogs))
        else:
            logging.info(f"No word types exceed {ratio}% limit.")


def strLen(items):
    return str(len(items))


class ResultLogger():

    def __init__(self, shh, letterList, leafIndex, newCache):
        c0, c1 = newCache.counts
        lengths = list(map(strLen, letterList))
        variables = lengths[c0:]
        prod = "x".join(variables)
        self.debugList = [
            f'Seached <{prod} sentences"',
            f"From {c0}w cache to {c1}w cache",
        ]
        self.logList = []
        self.tries = 0
        self.shh = shh

    def add(self, item):
        self.logList.append(item)

    def log(self, leaf, leafIndex):
        debugList = self.debugList
        logList = self.logList
        tries = self.tries
        logLen = len(logList)
        if logLen:
            logging.info(f"Pattern #{1 + leafIndex}, ")
            if not self.shh:
                debugLog = " ".join(debugList + logList[:1])
                logging.info(f"Found {logLen}/{tries}, ")
                logging.debug(debugLog + ", ")
                logging.info("")



class Cache():

    def __init__(self, items, counts):
        self.counts = counts
        self.items = items

    def __iter__(self):
        for i, item in enumerate(self.items):
            yield (i, item)

    def clearEmptyAndUncertain(self, cleared):
        isEmpty = self.isEmpty
        isCertain = self.isCertain
        if isEmpty and not isCertain:
            self.items = cleared

    def updateCounts(self, c0, c1):
        self.counts = (c0, c1)
        self.clearEmptyAndUncertain([(0,)])

    @property
    def isEmpty(self):
        return len(self.items) == 0

    @property
    def isCertain(self):
        c0, c1 = self.counts
        return c0 <= c1

    @property
    def latest(self):
        return (self.items or [()])[-1]

    def toNew(self):
        counts = self.counts
        return Cache([], counts)

    def isEnough(self, count):
        c1 = self.counts[1]
        return count >= c1

    def addItem(self, indexList):
        c1 = self.counts[1]
        caching = tuple(indexList[:c1])
        if self.latest != caching:
            self.items.append(tuple(caching))



import logging

from .recounter import Recounter
from .cache import Cache


def compareLeaves(a, b):
    aList = a.split(' ')
    bList = b.split(' ')
    for i, (a, b) in enumerate(zip(aList, bList)):
        if a != b:
            return i + 1
    return 1


def toPangrams(maps, patterns, config):

    cachedLeaf = ''
    leafList = patterns.list
    cache = Cache([(0,)], (1, 0))
    logging.info(f'Using {len(leafList)} patterns')

    for leafIndex, (leaf, nextLeaf) in enumerate(zip(leafList, leafList[1:])):
        try:
            recounter = Recounter(maps, config, leaf)
        except ValueError as e:
            logging.error(f'impossible {e} in leaf: {leaf}')
            continue

        cacheC = compareLeaves(cachedLeaf, leaf)
        nextC = compareLeaves(leaf, nextLeaf)
        cache.updateCounts(cacheC, nextC)

        if not cache.isEmpty:
            cachedLeaf = leaf
            cache = recounter.guessNext(cache, leafIndex)

    return config



from .logging import ResultLogger

from itertools import product


def findCopies(keyList):
    keyEqualities = [a == b for a, b in zip(keyList[1:], keyList)]
    return [i + 1 for i, x in enumerate(keyEqualities) if x]


def prettyPrint(wnl):
    w, n, letters = wnl
    return w + ':' + '/'.join(map(str, [n, len(letters)]))


def makeIndexList(ivLength, cacheItem):
    numC = len(cacheItem)

    def cacheReader(i):
        return cacheItem[i] if i < numC else 0

    return list(map(cacheReader, range(ivLength)))


def yieldWords(wordMap, ivList, showAll):
    letterList = [iv['combo'] for iv in ivList]
    wordLists = [wordMap.get(c, ['']) for c in letterList]
    if showAll:
        for words in product(*wordLists):
            yield tuple(words)
    else:
        yield tuple([words[0] for words in wordLists])


class Recounter:

    def __init__(self, maps, config, leaf):

        letterMap = maps.letterMap
        keyList = [""] + leaf.split(" ")
        self.letterList = [letterMap[k] for k in keyList]
        self.wordMap = maps.wordMap
        self.length = len(keyList)
        self.keyList = keyList
        self.config = config
        self.leaf = leaf

    def addOutput(self, words):
        self.config.addOutput(words)

    def addEmpty(self, line):
        self.config.addEmpty(line)

    def recount(self, cacheItem):
        self.combIndexList = makeIndexList(self.length, cacheItem)
        zipped = zip(cacheItem, self.letterList, self.keyList)
        addedStart = self.combiner(list(zipped)[1:], 0)
        ivBase = [{'combo': '', 'integer': 0}]
        self.ivListRoot = ivBase + addedStart
        self.copiedKeys = findCopies(self.keyList)
        self.pivot = len(cacheItem)

    def skipper(self, valid):
        return max(0, self.length - valid - 1)

    def stringifyWordList(self, words):
        letterList = self.letterList
        combIndexList = self.combIndexList
        iter = zip(words, combIndexList, letterList)
        return ' '.join(list(map(prettyPrint, iter))[1:])

    def logNexts(self, words):
        outCount = self.config.outputLength
        stats = self.stringifyWordList(words)
        return f'Result #{outCount}: {stats}'

    def iterateIteration(self, skipping):
        newIndexList = [0]*skipping
        letterList = self.letterList
        combIndexList = self.combIndexList
        iter = list(zip(combIndexList, letterList))
        for n, l in iter[::-1][skipping:]:
            nexts = n + 1
            count = len(l)
            if nexts < count:
                newIndexList = [nexts] + newIndexList
                break
            newIndexList = [0] + newIndexList

        if len(combIndexList) == len(newIndexList):
            return [1] + newIndexList[1:]
        pivot = len(combIndexList) - len(newIndexList)
        return combIndexList[:pivot] + newIndexList

    def handleNext(self, logger, newCache):

        copiedKeys = self.copiedKeys

        ivList = self.ivListRoot + self.combine()
        isValid, skipping = self.verify(ivList)

        # Ensure no duplicates among identical keys
        # Identical keys like "3x" are possible
        for dup in copiedKeys:
            skip = self.skipper(dup)
            if skip >= skipping:
                iv0 = self.combIndexList[dup - 1]
                iv1 = self.combIndexList[dup]
                if iv0 <= iv1:
                    self.combIndexList = self.iterateIteration(skip)
                    return newCache

        # Cache partial matches
        if newCache.isEnough(len(ivList)):
            newCache.addItem(self.combIndexList)

        # Update
        if isValid:
            showAll = self.config.showAll
            for words in yieldWords(self.wordMap, ivList, showAll):
                logger.add(self.logNexts(words))
                self.addOutput(words)

        # Iterate
        self.combIndexList = self.iterateIteration(skipping)
        logger.tries += 1
        return newCache

    def guessNext(self, cache, leafIndex):
        shh = self.config.shh
        newCache = cache.toNew()
        letterList = self.letterList
        logArgs = (letterList, leafIndex, newCache)
        logger = ResultLogger(shh, *logArgs)

        for i, cacheItem in cache:

            self.recount(cacheItem)

            while self.notComplete(cacheItem):
                newCache = self.handleNext(logger, newCache)

        if not shh or leafIndex % 100 == 0:
            logger.log(self.leaf, leafIndex)
            if len(logger.logList) < 0:
                self.addEmpty(self.leaf)

        return newCache

    def notComplete(self, cacheItem):
        lastCache = cacheItem[-1]
        cacheLength = len(cacheItem) - 1
        if self.combIndexList[0] != 0:
            return False
        if self.combIndexList[cacheLength] != lastCache:
            return False
        return True

    def verify(self, ivList):
        valid = len(ivList)
        isValid = valid == self.length
        return isValid, self.skipper(valid)

    def maskCombo(self, c):
        bits = ""
        for a in self.config.cList:
            bits += "1" if a in c else "0"
        return int(bits, 2)

    def combiner(self, zipped, allConsonants):
        combos = []
        for next, letters, key in zipped:
            if next >= len(letters):
                break
            combo = list(letters[next])
            integer = self.maskCombo(combo)
            if (allConsonants & integer) != 0:
                break
            allConsonants |= integer
            out = {
                "combo": ''.join(combo),
                "integer": allConsonants
            }
            combos.append(out)
        return combos

    def combine(self):
        keyList = self.keyList
        letterList = self.letterList
        combIndexList = self.combIndexList
        allConsonants = int(self.ivListRoot[-1]["integer"])
        zipped = list(zip(combIndexList, letterList, keyList))[self.pivot:]
        return self.combiner(zipped, allConsonants)



from .logging import logMaxCounts
from .flushHandler import FlushHandler

from statistics import variance
from pathlib import Path
import logging
import math


class Config:

    cList = 'bcdfghjklmnpqrstvwxz'
    vList = 'aeiouy'

    def __init__(self, shh, showAll, *args):
        (maxRatio, maxCounts, maxReps, priors) = args[:4]
        (minWords, maxWords, maxC, maxV) = args[4:]
        self.flushHandler = FlushHandler()
        self.maxRatio = maxRatio
        self.maxCounts = maxCounts
        self.maxReps = maxReps
        self.priors = priors
        self.minWords = minWords
        self.maxWords = maxWords
        self.maxConsonants = maxC
        self.maxVowels = maxV
        self.output = []
        self.source = []
        self.empty = []
        self.showAll = showAll
        self.shh = shh

    def makeFolder(self, folder):
        isDir = folder.is_dir()
        verb = "Using" if isDir else "Making"
        logFn = logging.info if isDir else logging.warning
        folder.mkdir(parents=True, exist_ok=True)
        logFn(f"{verb} {folder}")

    def makeFolders(self, parents):
        prefix = Path(self.prefix)
        for parent in parents:
            self.makeFolder(parent / prefix)
        testPath = self.toPath("*", "ext")
        logging.info(f"Paths like: {testPath}")

    def toPath(self, name, ext):
        prefix = Path(self.prefix)
        suffix = self.suffix
        text = "_".join([name, suffix])
        return prefix / Path(f"{text}.{ext}")

    def setEmpty(self, empty):
        self.empty = empty

    def addEmpty(self, line):
        self.empty.append(line)
        self.empty = self.flushHandler.flushEmpty(self)

    def addSource(self, wf):
        self.source.append(wf)

    def addOutput(self, words):
        self.output.append(" ".join(words))
        self.output = self.flushHandler.flushOutput(self)

    @property
    def outputLength(self):
        return len(self.output)

    @property
    def prefix(self):
        maxReps = self.maxReps
        maxRatio = self.maxRatio
        pVals = self.priors.values()
        priorValues = pVals if len(pVals) > 2 else [1, 1]
        std = math.sqrt(variance(priorValues))
        std100 = math.ceil(100 * std)
        max100 = int(100 * maxRatio)
        limitList = [k for k, v in maxReps.items() if v > 0]
        limited = "_".join(sorted(limitList))
        prefix = f"_has_{limited}" if len(limited) else ""
        return f'std{std100}_max{max100}{prefix}'

    @property
    def suffix(self):
        maxCounts = self.maxCounts
        mk = maxCounts.keys()
        if "*" not in maxCounts:
            raise ValueError(logMaxCounts(maxCounts))
        mk = sorted([k for k in mk if isinstance(k, int)])
        counts = "_".join([f"{maxCounts[k]}x{k}" for k in mk])
        suffix = f"_{counts}" if len(counts) else ""
        return f'{maxCounts["*"]}{suffix}'

    def wordRange(self):
        minWords = self.minWords
        maxWords = self.maxWords
        wordBounds = (minWords, maxWords + 1)
        return list(range(*wordBounds))

    def vowelCombos(self):
        mV = len(self.vList)
        nV = range(self.maxVowels)
        return sum([math.comb(mV, n) for n in nV])



from google_ngram_downloader import readline_google_store
import unidecode


def isWordOkay(word):
    for part in word.split("_"):
        isText = all([c.isalpha() for c in part])
        isReal = any([not c.isupper() for c in part])
        if not isText or not isReal:
            return False
    return True


def readWordRecords():
    lists = readline_google_store(ngram_len=1)
    for fname, url, records in lists:
        filename = fname.split('.')[0]
        suffix = filename.split('-')[-1]
        if len(suffix) > 1 or not suffix[0].isalpha():
            continue
        for record in records:
            if record.year == 2008:
                word = unidecode.unidecode(record.ngram)
                if isWordOkay(word):
                    frequency = record.match_count
                    yield [word.lower(), frequency]



import logging

class FlushHandler():

    EMPTY_MAX = 100
    OUTPUT_MAX = 1000

    def __init__(self, flushEmpty=None, flushOutput=None):
        self._flushEmpty = flushEmpty
        self._flushOutput = flushOutput

    def flushEmpty(self, config):
        empty = config.empty
        emptyCount = len(empty)
        if self._flushEmpty and emptyCount >= self.EMPTY_MAX:
            logging.info(f'Flushing {emptyCount} empty patterns.')
            self._flushEmpty(config)
            return []
        return empty

    def flushOutput(self, config):
        output = config.output
        outputCount = len(output)
        if self._flushOutput and len(output) >= self.OUTPUT_MAX:
            logging.info(f'Flushing {outputCount} output patterns.')
            self._flushOutput(config)
            return []
        return output



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

classifiers = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Natural Language :: English",
    "Intended Audience :: Science/Research",
    "Topic :: Text Processing :: Linguistic",
    "Topic :: Scientific/Engineering :: Information Analysis",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

setuptools.setup(
    name="abeci",
    version="1.0.5",
    author="TVQuizPhD",
    long_description=long_description,
    author_email="tvquizphd@email.com",
    description="Find perfect pangrams",
    url="https://github.com/tvquizphd/abeci",
    long_description_content_type="text/markdown",
    entry_points=dict(console_scripts=[
        'abeci-pangrams = abeci.pangrams:pangramsCli'
    ]),
    packages=setuptools.find_packages('src'),
    package_data={'abeci': ['typo/data/*.txt']},
    package_dir={'':'src'},
    install_requires=['abeci'],
    classifiers=classifiers
)



def flattenList(tree):
    if len(tree) == 0:
        return tree
    if isinstance(tree[0], list):
        return flattenList(tree[0]) + flattenList(tree[1:])
    return tree[:1] + flattenList(tree[1:])


def toLeafList(tree, empty):
    flat = flattenList(tree)
    leaves = [i for i in flat if i not in empty]
    return leaves + ['']


class Patterns():

    def __init__(self, tree, empty):
        self.tree = tree
        self.list = toLeafList(tree, empty)


def patternFunction(empty):
    def toPatterns(tree):
        return Patterns(tree, empty)
    return toPatterns



import re
'''
    Commonly Misspelled English words stored into json.
    Corpora attained from Roger Milton's hompage
    https://www.dcs.bbk.ac.uk/~ROGER/corpora.html
'''
pattern = re.compile(r'\<ERR(.*?)\</ERR>')


def loadTypos(dataDir):
    # Birbeck misspellings
    birkbeck_file = open(dataDir / 'missp.dat.txt')
    birkbeck_data = birkbeck_file.readlines()

    # Holbrook misspellings
    holbrook_file = open(dataDir / 'holbrook-tagged.dat.txt')
    holbrook_data = holbrook_file.readlines()

    # Aspell misspellings
    aspell_file = open(dataDir / 'aspell.dat.txt')
    aspell_data = aspell_file.readlines()

    # Wikipedia misspellings
    wikipedia_file = open(dataDir / 'wikipedia.dat.txt')
    wikipedia_data = wikipedia_file.readlines()

    misspellings = set()

    def store_misspellings(data, misspellings):
        for data_line in data:
            if '$' not in data_line:
                mis_spell = data_line.strip()
                misspellings.add(mis_spell.lower())

    for data_line in holbrook_data:
        for match in pattern.finditer(data_line):
            split = match.groups()[0].split('> ')
            misspell = split[1].strip().lower()

            if misspell not in misspellings:
                misspellings.add(misspell)

    store_misspellings(birkbeck_data, misspellings)
    store_misspellings(aspell_data, misspellings)
    store_misspellings(wikipedia_data, misspellings)

    return misspellings
